# -*- coding: utf-8 -*-

""" tests.test_metadata_construction

(Mostly) unit tests for
conda-content-trust/conda_content_trust/metadata_construction.py.

Run the tests this way:
    pytest tests/test_metadata_construction.py

⚠️ Note that these tests may require more dependencies than the codebase
   itself:
     - pytest
     - parameterize?

"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

# std libs
import copy
import os

# external dependencies
import pytest
import cryptography.exceptions # for InvalidSignature

# this codebase
from conda_content_trust.metadata_construction import *
from conda_content_trust.common import ( # these aren't already imported by metadata_construction
        keyfiles_to_bytes, keyfiles_to_keys, checkformat_key, is_a_signable,
        checkformat_delegating_metadata)
from conda_content_trust.signing import wrap_as_signable, sign_signable

# Some REGRESSION test data.
KEYPAIR_NAME = 'keytest_old'
PRIVATE_BYTES = b'\xc9\xc2\x06\r~\r\x93al&T\x84\x0bI\x83\xd0\x02!\xd8\xb6\xb6\x9c\x85\x01\x07\xdat\xb4!h\xf97'
PUBLIC_BYTES = b"\x01=\xddqIb\x86m\x12\xba[\xae'?\x14\xd4\x8c\x89\xcf\x07s\xde\xe2\xdb\xf6\xd4V\x1eR\x1c\x83\xf7"
PUBLIC_HEX = '013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7'
PKGMGR_PUBLIC_HEX = 'f46b5a7caa43640744186564c098955147daa8bac4443887bc64d8bfee3d3569'
# Signature is over b'123456\x067890' using key PRIVATE_BYTES.
SIGNATURE = b'\xb6\xda\x14\xa1\xedU\x9e\xbf\x01\xb3\xa9\x18\xc9\xb8\xbd\xccFM@\x87\x99\xe8\x98\x84C\xe4}9;\xa4\xe5\xfd\xcf\xdaau\x04\xf5\xcc\xc0\xe7O\x0f\xf0F\x91\xd3\xb8"\x7fD\x1dO)*\x1f?\xd7&\xd6\xd3\x1f\r\x0e'
HASHED_VAL = b'string to hash\n'
HASH_HEX = '73aec9a93f4beb41a9bad14b9d1398f60e78ccefd97e4eb7d3cf26ba71dbe0ce'
#HASH_BYTES = b's\xae\xc9\xa9?K\xebA\xa9\xba\xd1K\x9d\x13\x98\xf6\x0ex\xcc\xef\xd9~N\xb7\xd3\xcf&\xbaq\xdb\xe0\xce'
REPODATA_HASHMAP = {
    "noarch/current_repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "noarch/repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "noarch/repodata_from_packages.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "osx-64/current_repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
    "osx-64/repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
    "osx-64/repodata_from_packages.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2"
}
TEST_TIMESTAMP = '2019-10-01T00:00:00Z'
TEST_EXPIRY_DATE = '2025-01-01T10:30:00Z'
EXPECTED_UNSIGNED_REPODATA_VERIFY = {
    'type': 'repodata_verify',
    'timestamp': TEST_TIMESTAMP,
    'metadata_spec_version': '0.1.0',
    'expiration': TEST_EXPIRY_DATE,
    'secured_files': {
        'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
}
# EXPECTED_REGSIGNED_REPODATA_VERIFY = {
#     # Re-sign this if its data changes: it's signed!
#     'type': 'repodata_verify', 'timestamp': '2019-10-01T00:00:00Z',
#     'metadata_spec_version': '0.1.0', 'expiration': '2025-01-01T10:30:00Z',
#     'secured_files': {
#         'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
# }
ROOT_PUBLIC_HEX = 'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'
EXPECTED_UNSIGNED_ROOT = {
    'type': 'root',
    'timestamp': TEST_TIMESTAMP,
    'version': 1,
    'metadata_spec_version': '0.6.0',
    'expiration': TEST_EXPIRY_DATE,
    'delegations': {
        'key_mgr': {
            'pubkeys': ['013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7'],
            'threshold': 1},
        'root': {
            'pubkeys': ['bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'],
            'threshold': 1}}
}




# def test_build_repodata_verification_metadata():
#     # Test only construction of (unsigned) repodata_verify.

#     # Regression
#     rd_v_md = build_repodata_verification_metadata(
#             REPODATA_HASHMAP,
#             expiry=TEST_EXPIRY_DATE,
#             timestamp=TEST_TIMESTAMP)
#     assert rd_v_md == EXPECTED_UNSIGNED_REPODATA_VERIFY

#     # Bad-argument tests, expecting TypeErrors
#     bad_hashmap = copy.deepcopy(REPODATA_HASHMAP)
#     bad_hashmap['some_filename'] = 'this is not a hash'

#     with pytest.raises(ValueError):
#         build_repodata_verification_metadata(bad_hashmap)
#     with pytest.raises(ValueError):
#         build_repodata_verification_metadata(5) # not a hashmap at all


#     assert not is_a_signable(rd_v_md)
#     signable_rd_v_md = wrap_as_signable(rd_v_md)
#     assert is_a_signable(signable_rd_v_md)

#     private = PrivateKey.from_bytes(PRIVATE_BYTES)

#     sign_signable(signable_rd_v_md, private)
#     assert is_a_signable(signable_rd_v_md)




def test_build_root_metadata():
    # Test only construction of (unsigned) root metadata.

    # Regression
    root_md = build_root_metadata(
            root_pubkeys=[ROOT_PUBLIC_HEX],
            root_threshold=1,
            root_version=1,
            root_expiration=TEST_EXPIRY_DATE,
            key_mgr_pubkeys=[PUBLIC_HEX],
            key_mgr_threshold=1,
            root_timestamp=TEST_TIMESTAMP)

    assert root_md == EXPECTED_UNSIGNED_ROOT

    # This format check expects a signing envelope.
    checkformat_delegating_metadata(wrap_as_signable(root_md))


    # # Bad-argument tests, expecting TypeErrors
    # bad_hashmap = copy.deepcopy(REPODATA_HASHMAP)
    # bad_hashmap['some_filename'] = 'this is not a hash'

    # Bad-argument tests, expecting TypeErrors or ValueErrors
    with pytest.raises(ValueError):
        root_md = build_root_metadata(
                root_pubkeys=[ROOT_PUBLIC_HEX[:-1]],  # too short to be a key
                root_threshold=1,
                root_version=1,
                root_expiration=TEST_EXPIRY_DATE,
                key_mgr_pubkeys=[PUBLIC_HEX],
                key_mgr_threshold=1,
                root_timestamp=TEST_TIMESTAMP)

    with pytest.raises(TypeError):
        root_md = build_root_metadata(
                root_pubkeys=[ROOT_PUBLIC_HEX],
                root_threshold='this is not an integer', #  <---
                root_version=1,
                root_expiration=TEST_EXPIRY_DATE,
                key_mgr_pubkeys=[PUBLIC_HEX],
                key_mgr_threshold=1,
                root_timestamp=TEST_TIMESTAMP)

    with pytest.raises(ValueError):
        root_md = build_root_metadata(
                root_pubkeys=ROOT_PUBLIC_HEX,  # not a list of keys
                root_threshold=1,
                root_version=1,
                root_expiration=TEST_EXPIRY_DATE,
                key_mgr_pubkeys=[PUBLIC_HEX],
                key_mgr_threshold=1,
                root_timestamp=TEST_TIMESTAMP)

    with pytest.raises(ValueError):
        root_md = build_root_metadata(
                root_pubkeys=[ROOT_PUBLIC_HEX],
                root_threshold=1,
                root_version='this is not a version number',
                root_expiration=TEST_EXPIRY_DATE,
                key_mgr_pubkeys=[PUBLIC_HEX],
                key_mgr_threshold=1,
                root_timestamp=TEST_TIMESTAMP)

    with pytest.raises(TypeError):
        root_md = build_root_metadata(
                root_pubkeys=[ROOT_PUBLIC_HEX],
                root_threshold=1,
                root_version=1,
                root_expiration=91,             # <------
                key_mgr_pubkeys=[PUBLIC_HEX],
                key_mgr_threshold=1,
                root_timestamp=TEST_TIMESTAMP)

    assert not is_a_signable(root_md)
    signable_root_md = wrap_as_signable(root_md)
    assert is_a_signable(signable_root_md)


def test_build_delegating_metadata():
    # See also test_build_root_metadata.

    key_mgr = build_delegating_metadata(
            metadata_type='key_mgr', # 'root' or 'key_mgr'
            delegations={'pkg_mgr': {
                'pubkeys': [PKGMGR_PUBLIC_HEX],
                'threshold': 1}},
            version=1,
            #timestamp   default: now
            #expiration  default: now plus root expiration default duration
            )

    key_mgr = wrap_as_signable(key_mgr)

    checkformat_delegating_metadata(key_mgr)



def test_gen_and_write_keys():

    # Make a new keypair.  Returns keys and writes keys to disk.
    # Then load it from disk and compare that to the return value.  Exercise
    # some of the functions redundantly.
    try:
        generated_private, generated_public = gen_and_write_keys('keytest_new')
        loaded_new_private_bytes, loaded_new_public_bytes = keyfiles_to_bytes(
                'keytest_new')
        loaded_new_private, loaded_new_public = keyfiles_to_keys('keytest_new')
        assert generated_private.is_equivalent_to(loaded_new_private)
        assert generated_public.is_equivalent_to(loaded_new_public)
        assert loaded_new_private.is_equivalent_to(
                    PrivateKey.from_bytes(loaded_new_private_bytes))
        assert loaded_new_public.is_equivalent_to(
                    PublicKey.from_bytes(loaded_new_public_bytes))

    finally:
        # Clean files up.
        for fname in [
                'keytest_new.pub', 'keytest_new.pri',
                'keytest_old.pri', 'keytest_old.pub']:
            if os.path.exists(fname):
                os.remove(fname)


    # TODO: ✅ Some more tests are warranted.

    # # Clean variables up a bit for the next tests.
    # new_private = loaded_new_private
    # new_public = loaded_new_public
    # # old_private = loaded_old_private
    # # old_public = loaded_old_public
    # del (
    #         loaded_new_public, loaded_new_private,
    #         # loaded_old_private, loaded_old_public,
    #         generated_private, generated_public,
    #         loaded_new_private_bytes, loaded_new_public_bytes)



def test_gen_keys():
    # Note that test_authentication uses gen_keys to test creation of new
    # keys and their use in signing and verification.

    generated_private_1, generated_public_1 = gen_keys()
    generated_private_2, generated_public_2 = gen_keys()

    checkformat_key(generated_private_1)
    checkformat_key(generated_public_1)
    checkformat_key(generated_private_2)
    checkformat_key(generated_public_2)

    assert not generated_private_1.is_equivalent_to(generated_private_2)
    assert not generated_private_1.is_equivalent_to(generated_public_1)
    assert not generated_public_1.is_equivalent_to(generated_public_2)

    sig_from_1 = generated_private_1.sign(b'1234')
    sig_from_2 = generated_private_2.sign(b'1234')

    generated_public_1.verify(sig_from_1, b'1234')
    generated_public_2.verify(sig_from_2, b'1234')

    with pytest.raises(cryptography.exceptions.InvalidSignature):
        generated_public_1.verify(sig_from_2, b'1234')
        generated_public_1.verify(sig_from_1, b'5678')





# -*- coding: utf-8 -*-

""" conda_content_trust.root_signing
This module contains functions that sign data in an OpenPGP-compliant (i.e.
GPG-friendly) way.  Root metadata may be signed in this manner.  Functions that
perform simpler, direct signing using raw ed25519 keys are provided in
conda_content_trust.signing instead.

This library takes advantage of the securesystemslib library for its gpg
signing interface.

Function Manifest for this Module:
    sign_via_gpg                  # requires securesystemslib
    sign_root_metadata_via_gpg    # requires securesystemslib
    fetch_keyval_from_gpg         # requires securesystemslib

These two functions are provided only for testing purpose and are not part of
the API for this module:
    _gpg_pubkey_in_ssl_format
    _verify_gpg_sig_using_ssl     # requires securesystemslib

Note that there is a function in conda_content_trust.authentication that verifies these
signatures without requiring securesystemslib.
"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

# std libs
import json
#import binascii  # for binascii.unhexlify / hexlify
#import struct    # for struct.pack

# dependencies
# For ed25519 signing operations and hashing
import cryptography.hazmat.primitives.asymmetric.ed25519# as ed25519
import cryptography.hazmat.primitives.hashes# as pyca_crypto_hashes
import cryptography.hazmat.backends# as pyca_crypto_backends
import cryptography.exceptions# as pyca_crypto_exceptions

# securesystemslib is an optional dependency, and required only for signing
# root metadata via GPG.  Verification of those signatures, and signing other
# metadata with raw ed25519 signatures, does not require securesystemslib.
try:
    import securesystemslib.gpg.functions as gpg_funcs
    import securesystemslib.formats
    SSLIB_AVAILABLE = True
except ImportError:
    SSLIB_AVAILABLE = False

# this codebase
from .common import (
        canonserialize, load_metadata_from_file, write_metadata_to_file,
        is_a_signable,
        checkformat_gpg_fingerprint, checkformat_hex_key,
        checkformat_gpg_signature, checkformat_byteslike,
        PrivateKey, PublicKey, checkformat_key)



def sign_via_gpg(data_to_sign, gpg_key_fingerprint, include_fingerprint=False):
    """
    <Purpose>

        This is an alternative to the conda_content_trust.common.PrivateKey.sign() method, for
        use with OpenPGP keys, allowing us to use protected keys in YubiKeys
        (which provide an OpenPGP interface) to sign data.

        The signature is not simply over data_to_sign, as is the case with the
        PrivateKey.sign() function, but over an expanded payload with
        metadata about the signature to be signed, as specified by the OpenPGP
        standard (RFC 4880).  See data_to_sign and Security Note below.

        This process is nominally deterministic, but varies with the precise
        time, since there is a timestamp added by GPG into the signed payload.
        Nonetheless, this process does not depend at any point on the ability
        to generate random data (unlike key generation).

        This function requires securesystemslib, which is otherwise an optional
        dependency.

    <Arguments>

        data_to_sign
            The raw bytes of interest that will be signed by GPG.  Note that
            pursuant to the OpenPGP standard, GPG will add to this data:
            specifically, it includes metadata about the signature that is
            about to be made into the data that will be signed.  We do not care
            about that metadata, and we do not want to burden signature
            verification with its processing, so we essentially ignore it.
            This should have negligible security impact, but for more
            information, see "A note on security" below.


        gpg_key_fingerprint
            This is a (fairly) unique identifier for an OpenPGP key pair.
            Also Known as a "long" GPG keyid, a GPG fingerprint is
            40-hex-character string representing 20 bytes of raw data, the
            SHA-1 hash of a collection of the GPG key's properties.
            Internally, GPG uses the key fingerprint to identify keys the
            client knows of.

            Note that an OpenPGP public key is a larger object identified by a
            fingerprint.  GPG public keys include two things, from our
            perspective:

             - the raw bytes of the actual cryptographic key
               (in our case the 32-byte value referred to as "q" for an ed25519
               public key)

             - lots of data that is totally extraneous to us, including a
               timestamp, some representations of relationships with other keys
               (subkeys, signed-by lists, etc.), Superman's real name
               (see also https://bit.ly/38GcaGj), potential key revocations,
               etc.
               We do not care about this extra data because we are using the
               OpenPGP standard not for its key-to-key semantics or any element
               of its Public Key Infrastructure features (revocation, vouching
               for other keys, key relationships, etc.), but simply as a means
               of asking YubiKeys to sign data for us, with ed25519 keys whose
               raw public key value ("q") we know to expect.


    <Returns>
        Returns a dictionary representing a GPG signature.  This is similar to
        but not *quite* the same as
        securesystemslib.formats.GPG_SIGNATURE_SCHEMA (which uses 'keyid'
        as the key for the fingerprint, instead of 'gpg_key_fingerprint').

        Specifically, this looks like:
            {'gpg_key_fingerprint': <gpg key fingerprint>,
            'other_headers':       <extra data mandated in OpenPGP signatures>,
            'signature':        <ed25519 signature, 64 bytes as 128 hex chars>}


        This is unlike conda_content_trust.signing.sign(), which simply returns 64 bytes of raw
        ed25519 signature.


    <Security Note>

        A note on the security implications of this treatment of OpenPGP
        signatures:

        TL;DR:
            It is NOT easier for an attacker to find a collision; however, it
            IS easier, IF an attacker CAN find a collision, to do so in a way
            that presents a specific, arbitrary payload.

        Note that pursuant to the OpenPGP standard, GPG will add to the data we
        ask it to sign (data_to_sign) before signing it. Specifically, GPG will
        add, to the payload-to-be-signed, OpenPGP metadata about the signature
        it is about to create.  We do not care about that metadata, and we do
        not want to burden signature verification with its processing (that is,
        we do not want to use GPG to verify these signatures; conda will do
        that with simpler code).  As a result, we will ignore this data when
        parsing the signed payload.  This will mean that there will be many
        different messages that have the same meaning to us:

            signed:
                <some raw data we send to GPG: 'ABCDEF...'>
                <some data GPG adds in: '123456...'>

            Since we will not be processing the '123456...' above, '654321...'
            would have the same effect: as long as the signature is verified,
            we don't care what's in that portion of the payload.

        Since there are many, many payloads that mean the same thing to us, an
        attacker has a vast space of options all with the same meaning to us in
        which to search for (effectively) a useful SHA256 hash collision to
        find different data that says something *specific* and still
        *succeeds* in signature verification using the same signature.
        While that is not ideal, it is difficult enough simply to find a SHA256
        collision that this is acceptable.
    """
    if not SSLIB_AVAILABLE:
        # TODO✅: Consider a missing-optional-dependency exception class.
        raise Exception(
                'sign_via_gpg requires the securesystemslib library, which '
                'appears to be unavailable.')

    # Argument validation
    checkformat_gpg_fingerprint(gpg_key_fingerprint)
    checkformat_byteslike(data_to_sign)


    # try:
    #     full_gpg_pubkey = gpg_funcs.export_pubkey(gpg_key_fingerprint)
    # except securesystemslib.gpg.exceptions.KeyNotFoundError as e:
    #     raise Exception( # TODO✅: Consider an appropriate error class.
    #             'The GPG application reported that it is not aware of a key '
    #             'with the fingerprint provided ("' + str(gpg_key_fingerprint) +
    #             '").  You may need to import the given key.')

    sig = gpg_funcs.create_signature(data_to_sign, gpg_key_fingerprint)

    # # 💣💥 Debug only.
    # # 💣💥 Debug only.
    # assert gpg_funcs.verify_signature(sig, full_gpg_pubkey, data_to_sign)

    # securesystemslib.gpg makes use of the GPG key fingerprint.  We don't
    # care about that as much -- we want to use the raw ed25519 public key
    # value to refer to the key in a manner consistent with the way we refer to
    # non-GPG (non-OpenPGP) keys.
    keyval = fetch_keyval_from_gpg(gpg_key_fingerprint)

    # ssl gpg sigs look like this:
    #
    #   {'keyid': <gpg key fingerprint>,
    #    'other_headers': <extra data mandated in OpenPGP signatures>,
    #    'signature': <actual ed25519 signature, 64 bytes as 128 hex chars>}
    #
    # We want to store the real public key instead of just the gpg key
    # fingerprint, so we add that, and we'll rename keyid to
    # gpg_key_fingerprint.  That gives us:
    #
    #   {'gpg_key_fingerprint': <gpg key fingerprint>,
    #    'other_headers': <extra data mandated in OpenPGP signatures>,
    #    'signature': <actual ed25519 signature, 64 bytes as 128 hex chars>}
    #
    # sig['key'] = keyval  # q, the 32-byte raw ed25519 public key value, expressed as 64 hex characters

    # The OpenPGP Fingerprint of the OpenPGP key used to sign.  This is not
    # required for verification, but it's useful for debugging and for
    # root keyholder convenience.  So it's optional.
    if include_fingerprint:
        sig['see_also'] = sig['keyid'] # strictly not needed, useful for debugging; 20-byte sha1 gpg key identifier per OpenPGP spec, expressed as 40 hex characters

    del sig['keyid']

    return sig



# TODO✅: Rename this to sign_root_metadata_via_gpg and rename
#         the old sign_root_metadata_via_gpg to sign_root_metadata_file_via_gpg
def sign_root_metadata_dict_via_gpg(root_signable, gpg_key_fingerprint):
    # Signs root_signable in place, returns nothing.

    if not SSLIB_AVAILABLE:
        # TODO✅: Consider a missing-optional-dependency exception class.
        raise Exception(
                'sign_root_metadata_via_gpg requires the securesystemslib library, which '
                'appears to be unavailable.')


    # Make sure it's the right format.
    if not is_a_signable(root_signable):
        raise TypeError(
                'Expected a signable dictionary; the given file ' +
                str(root_md_fname) + ' failed the check.')
    # TODO: Add root-specific checks.

    # Canonicalize and serialize the data, putting it in the form we expect to
    # sign over.  Note that we'll canonicalize and serialize the whole thing
    # again once the signatures have been added.
    data_to_sign = canonserialize(root_signable['signed'])

    # sig_dict, pgp_pubkey = sign_via_gpg(data_to_sign, gpg_key_fingerprint)
    sig_dict = sign_via_gpg(data_to_sign, gpg_key_fingerprint)

    # sig_dict looks like this:
    #     {'keyid': 'f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589',
    #      'other_headers': '04001608001d162104f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd58905025dbc3e68',
    #      'signature': '29282a8fe75871f9d4cf10a5a9e8d92303f8c361ce4b474a0ce641c9b8a74e4baaf810cc383af318a8e21cbe252789c2c30894d94e8b0288c3c45ceacf6c1d0c'}
    # pgp_pubkey looks like this:
        # {'creation_time': 1571411344,
        # 'hashes': ['pgp+SHA2'],
        # 'keyid': 'f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589',
        # 'keyval': {'private': '',
        #            'public': {'q': 'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'}},
        # 'method': 'pgp+eddsa-ed25519',
        # 'type': 'eddsa'}

    # securesystemslib.gpg makes use of the GPG key fingerprint.  We don't
    # care about that as much -- we want to use the raw ed25519 public key
    # value to refer to the key in a manner consistent with the way we refer to
    # non-GPG (non-OpenPGP) keys.
    # raw_pubkey = pgp_pubkey['keyval']['public']['q']
    raw_pubkey = fetch_keyval_from_gpg(gpg_key_fingerprint)


    # non-GPG signing here would look like this:
    # signature_as_hexstr = serialize_and_sign(signable['signed'], private_key)
    # public_key_as_hexstr = binascii.hexlify(key_to_bytes(
    #         private_key.public_key())).decode('utf-8')


    # TODO: ✅⚠️ Log a warning in whatever conda's style is (or conda-build):
    #
    # if public_key_as_hexstr in signable['signatures']:
    #   warn(    # replace: log, 'warnings' module, print statement, whatever
    #           'Overwriting existing signature by the same key on given '
    #           'signable.  Public key: ' + public_key + '.')

    # Add signature in-place.
    root_signable['signatures'][raw_pubkey] = sig_dict

    return root_signable



def sign_root_metadata_via_gpg(root_md_fname, gpg_key_fingerprint):
    """
    # TODO✅: Proper docstring:
    # This is a higher-level function than sign_via_gpg, including code that
    # deals with the filesystem.  It is not actually limited to root metadata,
    # and SHOULD BE RENAMED.
    """

    # Read in json
    root_signable = load_metadata_from_file(root_md_fname)

    root_signable = sign_root_metadata_dict_via_gpg(root_signable, gpg_key_fingerprint)

    # TODO: Consider removing write_metadata_to_file.  It might be better for
    #       readers to see the canonserialize() call being made (again) here,
    #       and it's not that much longer....
    write_metadata_to_file(root_signable, root_md_fname)



def fetch_keyval_from_gpg(fingerprint):
    """
    Retrieve the underlying 32-byte raw ed25519 public key for a GPG key.

    Given a GPG key fingerprint (40-character hex string), retrieve the GPG
    key, parse it, and return "q", the 32-byte ed25519 key value.

    This takes advantage of the GPG key parser in securesystemslib.

    The fingerprint will be stripped of spaces and lowercased, so you can use
    the GPG output even if it's in a funky format:
            94A3 EED0 806C 1F10 7754  A446 FDAD 11B8 2DD4 0E8C
            94A3 EED0 806C 1F10 7754  A446 FDAD 11B8 2DD4 0E8C    # <-- No, this is actually not the same as the previous one, which uses \\xa0....
            94A3EED0806C1F107754A446FDAD11B82DD40E8C
            94a3eed0806c1f107754a446fdad11b82dd40e8c
            etc.
    """
    if not SSLIB_AVAILABLE:
        # TODO✅: Consider a missing-optional-dependency exception class.
        raise Exception(
                'fetch_keyval_from_gpg requires the securesystemslib library, which '
                'appears to be unavailable.')

    fingerprint = fingerprint.lower().replace(' ', '').replace('\xa0', '')  # \xa0 is another space character that GPG sometimes outputs

    checkformat_gpg_fingerprint(fingerprint)

    key_parameters = gpg_funcs.export_pubkey(fingerprint)

    return key_parameters['keyval']['public']['q']



def _verify_gpg_sig_using_ssl(signature, gpg_key_fingerprint, key_value, data):
    """
    THIS IS PROVIDED ONLY FOR TESTING PURPOSES.
    We will verify signatures using our own code in conda_content_trust.authentication, not
    by using the securesystemslib.gpg.functions.verify_signature call that
    sits here.

    Wraps securesystemslib.gpg.functions.verify_signature.  to format the
    arguments in a manner ssl will like (i.e. conforming to
    securesystemslib.formats.GPG_SIGNATURE_SCHEMA).
    """
    if not SSLIB_AVAILABLE:
        # TODO✅: Consider a missing-optional-dependency exception class.
        raise Exception(
                'verifygpg_sig_using_ssl requires the securesystemslib '
                'library, which appears to be unavailable.')

    checkformat_key(key_value)

    # This function validates these two args in the process of formatting them.
    ssl_format_key = gpg_pubkey_in_ssl_format(gpg_key_fingerprint, key_value)

    securesystemslib.formats.GPG_SIGNATURE_SCHEMA.check_match(signature)
    securesystemslib.formats._GPG_ED25519_PUBKEY_SCHEMA.check_match(
            ssl_format_key)



    # TODO: ✅ Validate sig (ssl-format gpg sig dict) and content (bytes).



    # Note: if we change the signature format to deviate from what ssl uses,
    #       then we need to correct it here if we're going to use ssl.




    validity = gpg_funcs.verify_signature(signature, ssl_format_key, data)

    return validity



def _gpg_pubkey_in_ssl_format(fingerprint, q):
    """
    THIS IS PROVIDED ONLY FOR TESTING PURPOSES.
    We do not need to convert pubkeys to securesystemslib's format, except to
    try out securesystemslib's gpg signature verification (which we use only
    for comparison during testing).

    Given a GPG key fingerprint (40 hex characters) and a q value (64 hex
    characters representing a 32-byte ed25519 public key raw value), produces a
    key object in a format that securesystemslib expects, so that we can use
    securesystemslib.gpg.functions.verify_signature for part of the GPG
    signature verification.  For our purposes, this means that we should
    produce a dictionary conforming to
    securesystemslib.formats._GPG_ED25519_PUBKEY_SCHEMA.

    If securesystemslib.formats._GPG_ED25519_PUBKEY_SCHEMA changes, those
    changes will likely need to be reflected here.

    Example value produced:
    {
        'type': 'eddsa',
        'method': 'pgp+eddsa-ed25519',
        'hashes': ['pgp+SHA2'],
        'keyid': 'F075DD2F6F4CB3BD76134BBB81B6CA16EF9CD589',
        'keyval': {
            'public': {'q': 'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'},
            'private': ''}
        }
    }
    """
    checkformat_gpg_fingerprint(fingerprint)
    checkformat_hex_key(q)

    ssl_format_key = {
        'type': 'eddsa',
        'method': securesystemslib.formats.GPG_ED25519_PUBKEY_METHOD_STRING,
        'hashes': [securesystemslib.formats.GPG_HASH_ALGORITHM_STRING],
        'keyid': fingerprint,
        'keyval': {'private': '', 'public': {'q': q}}
    }

    return ssl_format_key


# def _gpgsig_to_sslgpgsig(gpg_sig):
#
#     conda_content_trust.common.checkformat_gpg_signature(gpg_sig)
#
#     return {
#             'keyid': copy.deepcopy(gpg_sig['key_fingerprint']),
#             'other_headers': copy.deepcopy(gpg_sig[other_headers]),
#             'signature': copy.deepcopy(gpg_sig['signature'])}


# def _sslgpgsig_to_gpgsig(ssl_gpg_sig):
#
#     securesystemslib.formats.GPG_SIGNATURE_SCHEMA.check_match(ssl_gpg_sig)
#
#     return {
#             'key_fingerprint': copy.deepcopy(ssl_gpg_sig['keyid']),
#             'other_headers': copy.deepcopy(ssl_gpg_sig[other_headers]),
#             'signature': copy.depcopy(ssl_gpg_sig['signature'])
#     }



# -*- coding: utf-8 -*-

""" conda_content_trust.cli
This module provides the CLI interface for conda-content-trust.
This is intended to provide a command-line signing and metadata update
interface.
"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import json
from argparse import ArgumentParser
import copy
import json

from conda_content_trust.common import (
        canonserialize, load_metadata_from_file, write_metadata_to_file,
        CCT_Error, PrivateKey, is_gpg_fingerprint, is_hex_key)

from conda_content_trust import __version__
import conda_content_trust.root_signing as cct_root_signing
import conda_content_trust.signing as cct_signing
import conda_content_trust.authentication as cct_authentication
import conda_content_trust.metadata_construction as cct_metadata_construction

# In Python2, input() performs evaluation and raw_input() does not.  In
# Python3, input() does not perform evaluation and there is no raw_input().
# So... use raw_input in Python2, and input in Python3.
try:                     # pragma: no cover
    _input_func = raw_input
except NameError:        # pragma: no cover
    _input_func = input


def cli(args=None):
    p = ArgumentParser(
        description="Signing and verification tools for Conda",
        conflict_handler='resolve'
    )
    p.add_argument(
        '-V', '--version',
        action='version',
        help='Show the conda-content-trust version number and exit.',
        version="conda-content-trust %s" % __version__,
    )

    # Create separate parsers for the subcommands.
    sp = p.add_subparsers(title='subcommands', dest='subcommand_name')


    # subcommand: sign-artifacts

    p_signrepo = sp.add_parser(
            'sign-artifacts', help=('Given a repodata.json '
            'file, produce signatures over the metadata for each artifact listed, '
            'and update the repodata.json file with their individual signatures.'))
    p_signrepo.add_argument(
            'repodata_fname', help=('the filename of a repodata.json file from '
            'which to retrieve metadata for individual artifacts.'))
    p_signrepo.add_argument(
            'private_key_hex', help=('the ed25519 private key to be used to '
            'sign each artifact\'s metadata'))


    # subcommand: verify-metadata

    p_verifymd = sp.add_parser(
            'verify-metadata', help=('Uses the first (trusted) metadata file '
            'to verify the second (not yet trusted) metadata file.  For '
            'example, '
            '"conda-content-trust verify-metadata 4.root.json 5.root.json"'
            ' to verify version 5 of root based on version 4 of root, or '
            '"conda-content-trust verify-metadata 4.root.json key_mgr.json" '
            'to verify key manager metadata based on version 4 of root.'))
    p_verifymd.add_argument(
            'trusted_metadata_filename', help=('the filename of the '
            'already-trusted metadata file that sets the rules for verifying '
            'the untrusted metadata file'))
    p_verifymd.add_argument(
            'untrusted_metadata_filename', help=('the filename of the '
            '(untrusted) metadata file to verify'))


    # subcommand: modify-metadata

    p_modifymd = sp.add_parser(
            'modify-metadata', help=('Interactive metadata modification.  Use '
            'this to produce a new version of a metadata file (like root.json '
            'or key_mgr.json), or correct an error in an unpublished metadata '
            'file, or review and sign a metadata file.  This increments '
            'version number / timestamp, reports changes on console, etc. For '
            'example, "conda-content-trust modify-metadata 8.root.json" '
            'for assistance in '
            'producing a new version of root (version 9) using version 8.'))
    p_modifymd.add_argument(
            'metadata_filename', help=('the filename of the existing metadata '
            'file to modify'))


    # If we're missing optional requirements for the next few options, note
    # that in their help strings.
    opt_reqs_str = ''
    if not cct_root_signing.SSLIB_AVAILABLE:
        opt_reqs_str = ('[Unavailable]: Requires optional '
                'dependencies: securesystemslib and gpg.  ')


    # subcommand: gpg-key-lookup
    p_gpglookup = sp.add_parser('gpg-key-lookup', help=(opt_reqs_str +
        'Given the OpenPGP fingerprint of an ed25519-type OpenPGP key, fetch '
        'the actual ed25519 public key value of the underlying key.'))
    p_gpglookup.add_argument(
        'gpg_key_fingerprint',
        help=('the 40-hex-character key fingerprint (long keyid) for the '
        'OpenPGP/GPG key that you want to sign something with.  Do not '
        'add prefix "0x".'))


    # subcommand: gpg-sign

    p_gpgsign = sp.add_parser('gpg-sign', help=(opt_reqs_str + 'Sign a given '
        'piece of metadata using GPG instead of the usual signing '
        'mechanisms.  Takes an OpenPGP key fingerprint and a filename.'))
    p_gpgsign.add_argument(
        'gpg_key_fingerprint',
        help=('the 40-hex-character key fingerprint (long keyid) for the '
        'OpenPGP/GPG key that you want to sign something with.  Do not '
        'add prefix "0x".'))
    p_gpgsign.add_argument(
        'filename',
        help=('the filename of the file that will be signed'))



    args = p.parse_args(args)

    if args.subcommand_name == 'gpg-sign':

        # TODO: Validate arguments.

        # Strip any whitespace from the key fingerprint and lowercase it.
        # GPG pops out keys in a variety of whitespace arrangements and cases,
        # so this is necessary for convenience.
        gpg_key_fingerprint = ''.join(args.gpg_key_fingerprint.split()).lower()

        cct_root_signing.sign_root_metadata_via_gpg(
                args.filename, gpg_key_fingerprint)



    elif args.subcommand_name == 'sign-artifacts':

        cct_signing.sign_all_in_repodata(
                args.repodata_fname, args.private_key_hex)



    elif args.subcommand_name == 'gpg-key-lookup':
        gpg_key_fingerprint = ''.join(args.gpg_key_fingerprint.split()).lower()
        keyval = cct_root_signing.fetch_keyval_from_gpg(gpg_key_fingerprint)
        print('Underlying ed25519 key value: ' + str(keyval))


    elif args.subcommand_name == 'modify-metadata':

        # `conda-content-trust update-metadata <metadata file to produce new version of>`

        # underlying functions: build_delegating_metadata,
        # load_metadata_from_file

        # given a metadata file, increment the version number and timestamps,
        # reporting the changes on the console

        # strip signatures

        # indicate what signatures are required

        # ask if the user wants to sign; query for the key hex or fname;
        # ideally, offer this functionality for both root and non-root keys.
        # For root metadata, we can (and should) also report which keys are
        # expected / still needed in order for the metadata to be verifiable
        # according to the old metadata and the new metadata

        old_metadata = load_metadata_from_file(args.metadata_filename)

        # new_metadata = cct_metadata_construction.interactive_modify_metadata(old_metadata)
        # if new_metadata is not None and new_metadata:
        #     write_metadata_to_file(new_metadata, args.metadata_filename)

        interactive_modify_metadata(old_metadata)



    elif args.subcommand_name == 'verify-metadata':


        # `conda-content-trust verify-metadata <trusted delegating metadata> <untrusted
        # metadata> <(optional) role name>`

        # underlying functions: cct_authentication.verify_delegation,
        # load_metadata_from_file

        # takes two metadata files, the first being a trusted file that should
        # provide the verification criteria (expected keys and expected number
        # of keys) for the second file.  This should support root-root
        # verification (root chaining as currently implemented in
        # conda-content-trust) and delegation from one metadata type to another
        # (e.g. root to key_mgr)

        # conveys to the user whether or not the file is trusted, and for what
        # role.  e.g., would convey that the first file is (e.g.) a root
        # metadata file, that it provides a delegation to <role name>, and that
        # the <untrusted metadata> file provides <role name> and is signed
        # appropriately based on what the root metadata file requires of that
        # delegation.


        untrusted_metadata = load_metadata_from_file(
                args.untrusted_metadata_filename)

        trusted_metadata = load_metadata_from_file(
                args.trusted_metadata_filename)

        # TODO✅: Argument validation via the check_format_* calls.

        metadata_type = untrusted_metadata['signed']['type']

        if metadata_type == 'root':
            # Verifying root has additional steps beyond verify_delegation.
            try:
                cct_authentication.verify_root(trusted_metadata, untrusted_metadata)
                print('Root metadata verification successful.')
                return 0 # success

            except CCT_Error as e:
                errorcode = 10
                errorstring = str(e)

        else:
            # Verifying anything other than root just uses verify_delegation
            # directly.
            try:
                cct_authentication.verify_delegation(
                        delegation_name=metadata_type,
                        untrusted_delegated_metadata=untrusted_metadata,
                        trusted_delegating_metadata=trusted_metadata)
                print('Metadata verification successful.')
                return 0 # success

            except CCT_Error as e:
                errorcode = 20
                errorstring = str(e)


        # We should only get here if verification failed.
        print(
                'Verification of untrusted metadata failed.  Metadata '
                'type was "' + metadata_type + '".  Error reads:\n  "'
                + errorstring + '"')
        return errorcode   # failure; exit code





    else:
        print('No command provided.  Please use  "conda-content-trust -h" for help.')




def interactive_modify_metadata(metadata):
    """
    """

    # Update version if there is a version.
    # Update timestamp if there is a timestamp.
    #
    # Show metadata contents ('signed') -- pprint?
    #    indicate updated version/timestamp
    #
    # Changes phase:
    #    Prompt to
    #       (m) modify a value, (a) add a new entry, (d) delete an entry,
    #       (r) revert to original, (f) finish and sign ((move on to signing
    #       prompts))
    #
    # Signing phase:
    #   Show metadata again, ask if metadata looks right
    #   Show what keys the original was signed by and ask if those should be
    #     the keys used for the new version.
    #        ((Later: if root, vet against contents of new and old root versions))
    #   Prompt for key (raw key file, raw key data, or gpg key fingerprint)
    #   Sign using the given key (gpg if gpg, else normal signing mechanism).
    #   Write (making sure not to overwrite, and -- if root -- making sure to
    #     prepend "<version>." to root.json file.

    initial_metadata = metadata
    metadata = copy.deepcopy(initial_metadata)

    try:
        import pygments
        import pygments.lexers
        import pygments.formatters
    except ImportError:
        print(
                'interactive modify-metadata mode employs pygments for syntax '
                'highlighting, if pygments is available.  pygments was not '
                'found, so the JSON contents will be... uglier than they '
                'would otherwise be.  If you would like syntax highlighting '
                'and prettier printing of JSON, you may install pygments.')
        pygments = None
        from pprint import pprint

    # Build the modification options and prompt.
    def promptfor(s):
        return _input_func(F_INSTRUCT + '\n----- Please provide ' + s + ENDC + ': ')

    def fn_write():
        fname = promptfor('a filename to save this metadata as')
        print('Writing to file....')
        write_metadata_to_file(metadata, fname)
        print('Modified metadata written!')
        return 1

    def fn_abort():
        # TODO✅: Ask to confirm.
        print(RED + BOLD + '\nAborting!\n' + ENDC)
        return 1

    def fn_addsig():
        if not cct_root_signing.SSLIB_AVAILABLE:
            print(F_OPTS + 'Signing.  ' + RED + 'Please ABORT (control-c) if '
                    'the metadata above is not EXACTLY what you want to sign!'
                    + ENDC)
        key = promptfor(
                'a key: either:\n     - a 40-character-hex-string GPG PUBLIC '
                'key fingerprint\n'
                '       for GPG keys (e.g. root YubiKeys), or \n     - a '
                '64-character-hex-string PRIVATE key value for normal '
                'keys.\n\n     Whitespace will be removed and characters will '
                'be lowercased.\n     Key')
        key = ''.join(key.split()).lower()

        if is_hex_key(key):
            private_key = PrivateKey.from_hex(key)
            cct_signing.sign_signable(metadata, private_key)
            print(F_OPTS + '\n\n--- Successfully signed!  Please save.' + ENDC)

        elif is_gpg_fingerprint(key):
            try:
                cct_root_signing.sign_root_metadata_dict_via_gpg(metadata, key)
            except:
                print(F_OPTS + '\n\n--- ' + RED + 'Signing FAILED.'
                        + F_OPTS + '  Do you have this key loaded in GPG on '
                        'this system?')
            else:
                print(F_OPTS + '\n\n--- Successfully signed!  Please save.' + ENDC)

        else:
            print(F_OPTS + RED + 'Unable to recognize key.  Please try again.'
                + ENDC)
        return 0
    def fn_remsig():
        return 0
    def fn_update():
        return 0
    def fn_adddel():
        return 0
    def fn_remdel():
        return 0
    def fn_thresh():
        delegation = promptfor('a delegation name (one of the entries in the'
                '\n     "delegations" dictionary in the metadata above).  '
                'This will\n     be the delegation whose threshold number of '
                'required keys we\n     will change.')
        if delegation not in metadata['signed']['delegations']:
            print(F_OPTS + '\n\n--- ' + RED + 'Unable to find that delegation.'
                    '  Please try again.' + ENDC)
            return 0

        new_thresh = promptfor('a new threshold value.  The current value is '
                 + str(metadata['signed']['delegations'][delegation]['threshold']))

        try:
            new_thresh = int(new_thresh)
            assert new_thresh >= 1
        except:
            print(F_OPTS + '\n--- ' + RED + 'Invalid value.  Expecting integer '
                    'greater than or equal to 1.  Please try again.' + ENDC)
            return 0

        metadata['signed']['delegations'][delegation]['threshold'] = new_thresh

        print(F_OPTS + '\n--- Threshold successfully updated.' + ENDC)

        return 0
    def fn_addkey():
        return 0
    def fn_remkey():
        return 0

    options = {
        0: [fn_write, 'Done: write and save metadata'],
        1: [fn_abort, 'Abort: discard changes -- abort without writing'],
        2: [fn_addsig, 'Add a signature (sign with a key you have)'],
        3: [fn_remsig, 'Remove a signature'],
        4: [fn_update, 'Update any top-level dictionary entry'],
        5: [fn_adddel, 'Add a delegation'],
        6: [fn_remdel, 'Remove a delegation'],
        7: [fn_thresh, 'Change the threshold number of keys for a delegation'],
        8: [fn_addkey, 'Add an authorized key to a delegation'],
        9: [fn_remkey, 'Remove an authorized key from a delegation']}


    option_text = (
            F_INSTRUCT + '\n--- Please choose an operation by entering its '
            'number\n' + ENDC)
    for index in options:
        option_text += ('    ' + F_LABEL + str(index) + ENDC + ': '
                + options[index][1] + ENDC + '\n')


    done = False
    while not done:

        print(F_OPTS + BOLD + '\n\n---------------------\n--- Current metadata:\n---------------------\n' + ENDC)

        if pygments is not None:
            formatted_metadata = json.dumps(metadata, sort_keys=True, indent=4)
            print(pygments.highlight(
                    formatted_metadata.encode('utf-8'),
                    pygments.lexers.JsonLexer(),
                    pygments.formatters.TerminalFormatter()))
        else:
            pprint(metadata)

        print(option_text)
        selected = _input_func(F_OPTS + 'Choice: ' + ENDC)
        try:
            selected = int(selected)
        except:
            print(RED + BOLD + '\nInvalid entry.  Try again.\n' + ENDC)
            continue
        if selected not in options:
            print(RED + BOLD + '\nInvalid entry.  Try again.\n' + ENDC)
            continue

        print(F_OPTS + '\nChose "' + options[selected][1] + '"' + ENDC)

        done = options[selected][0]() # Run the func associated with the option.


    ### Pull modified from debugging script
    ### Pull modified from debugging script
    ### Pull modified from debugging script



# Basic text formatting string constants
PINK = '\033[95m'
BLUE = '\033[94m'
CYAN = '\033[96m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
RED = '\033[91m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'

# Complete formats
F_LABEL = ENDC + UNDERLINE + BOLD + PINK
F_INSTRUCT = ENDC + BOLD + PINK
F_OPTS = ENDC + GREEN





if __name__ == '__main__':
    import sys
    exit_status = cli(sys.argv[1:])
    sys.exit(exit_status)



# -*- coding: utf-8 -*-

""" tests.test_cli

Integration tests for conda-content-trust/conda_content_trust/cli.py.

Run the tests this way:
    pytest tests/test_cli.py

"""

import subprocess

import conda_content_trust.cli

def test_cli_basics():
  assert not subprocess.call(['conda-content-trust', '-V'])
  assert not subprocess.call(['conda-content-trust', '--version'])
  assert not subprocess.call(['conda-content-trust', '--help'])

def test_that_all_calls_complete():
  assert not subprocess.call([
      'conda-content-trust', 'verify-metadata',
      'tests/testdata/1.root.json', 'tests/testdata/2.root.json'])
  assert not subprocess.call([
      'conda-content-trust', 'verify-metadata',
      'tests/testdata/1.root.json', 'tests/testdata/key_mgr.json'])



# def test_gpg_key_fingerprint():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()

# def test_():
#   raise NotImplementedError()




# -*- coding: utf-8 -*-

""" tests.test_signing

Unit tests for conda-content-trust/conda_content_trust/signing.py


 NOTE that much of the signing module is currently tested in
   test_authentication.py instead.  Some unit tests are missing.


Run the tests this way:
    pytest tests/test_signing.py

"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import os, os.path
import shutil
import copy

from conda_content_trust.common import *
from conda_content_trust.authentication import *
from conda_content_trust.signing import *

import pytest

# Some REGRESSION test data.
REG__KEYPAIR_NAME = 'keytest_old'
REG__PRIVATE_BYTES = b'\xc9\xc2\x06\r~\r\x93al&T\x84\x0bI\x83\xd0\x02!\xd8\xb6\xb6\x9c\x85\x01\x07\xdat\xb4!h\xf97'
REG__PRIVATE_HEX = 'c9c2060d7e0d93616c2654840b4983d00221d8b6b69c850107da74b42168f937'
REG__PUBLIC_BYTES = b"\x01=\xddqIb\x86m\x12\xba[\xae'?\x14\xd4\x8c\x89\xcf\x07s\xde\xe2\xdb\xf6\xd4V\x1eR\x1c\x83\xf7"
REG__PUBLIC_HEX = '013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7'
# Signature is over b'123456\x067890' using key REG__PRIVATE_BYTES.
REG__SIGNATURE = b'\xb6\xda\x14\xa1\xedU\x9e\xbf\x01\xb3\xa9\x18\xc9\xb8\xbd\xccFM@\x87\x99\xe8\x98\x84C\xe4}9;\xa4\xe5\xfd\xcf\xdaau\x04\xf5\xcc\xc0\xe7O\x0f\xf0F\x91\xd3\xb8"\x7fD\x1dO)*\x1f?\xd7&\xd6\xd3\x1f\r\x0e'
REG__HASHED_VAL = b'string to hash\n'
REG__HASH_HEX = '73aec9a93f4beb41a9bad14b9d1398f60e78ccefd97e4eb7d3cf26ba71dbe0ce'
#REG__HASH_BYTES = b's\xae\xc9\xa9?K\xebA\xa9\xba\xd1K\x9d\x13\x98\xf6\x0ex\xcc\xef\xd9~N\xb7\xd3\xcf&\xbaq\xdb\xe0\xce'
REG__REPODATA_HASHMAP = {
    "noarch/current_repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "noarch/repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "noarch/repodata_from_packages.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
    "osx-64/current_repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
    "osx-64/repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
    "osx-64/repodata_from_packages.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2"
}
REG__TEST_TIMESTAMP = '2019-10-01T00:00:00Z'
REG__TEST_EXPIRY_DATE = '2025-01-01T10:30:00Z'
REG__EXPECTED_UNSIGNED_REPODATA_VERIFY = {
    'type': 'repodata_verify', 'timestamp': REG__TEST_TIMESTAMP,
    'metadata_spec_version': '0.1.0', 'expiration': REG__TEST_EXPIRY_DATE,
    'secured_files': {
        'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
}
REG__EXPECTED_REGSIGNED_REPODATA_VERIFY = {
    # Re-sign this if its data changes: it's signed!
    'type': 'repodata_verify', 'timestamp': '2019-10-01T00:00:00Z',
    'metadata_spec_version': '0.1.0', 'expiration': '2025-01-01T10:30:00Z',
    'secured_files': {
        'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
        'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
        'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
}
REG__REPODATA_SAMPLE_FNAME = 'tests/repodata_sample.json'
REG__REPODATA_SAMPLE_TEMP_FNAME = 'tests/repodata_sample_temp.json'



# def test_serialize_and_sign():
#     raise(NotImplementedError()
#             '⚠️ These tests are currently implemented in '
#             'test_authentication.py instead.'))


# def test_sign_signable():
#     raise(NotImplementedError(
#             '⚠️ These tests are currently implemented in '
#             'test_authentication.py instead.'))

# def test_wrap_as_signable():
#     raise(NotImplementedError(
#             '⚠️ This function is tested in multiple modules, but '
#             'a unit test should be constructed from those tests.'))


def test_sign_all_in_repodata():

    public = PublicKey.from_hex(REG__PUBLIC_HEX)

    # Make a test copy of the repodata sample, since we're going to
    # update it.
    if os.path.exists(REG__REPODATA_SAMPLE_TEMP_FNAME):
        os.remove(REG__REPODATA_SAMPLE_TEMP_FNAME)
    shutil.copy(
            REG__REPODATA_SAMPLE_FNAME,
            REG__REPODATA_SAMPLE_TEMP_FNAME)

        # grab data and use it to compare to what we produce in a bit

    repodata = load_metadata_from_file(REG__REPODATA_SAMPLE_FNAME)

    sign_all_in_repodata(REG__REPODATA_SAMPLE_TEMP_FNAME, REG__PRIVATE_HEX)

    repodata_signed = load_metadata_from_file(REG__REPODATA_SAMPLE_TEMP_FNAME)

    # Ensure that the rest of repodata is unchanged.
    repodata_signed_stripped = copy.deepcopy(repodata_signed)
    del repodata_signed_stripped['signatures']
    assert repodata_signed_stripped == repodata
    del repodata_signed_stripped

    # Make sure there is one signature entry for every artifact entry, and no
    # mystery entries.
    assert repodata['packages'].keys() == repodata_signed['signatures'].keys()

    for artifact_name in repodata['packages']:
        # There's a signature "by" this key listed for every artifact.
        assert REG__PUBLIC_HEX in repodata_signed['signatures'][artifact_name]
        # The signature is valid.
        verify_signature(
                # signature, (supposed) key, data that was (supposedly) signed
                repodata_signed['signatures'][artifact_name][REG__PUBLIC_HEX]['signature'],
                public,
                canonserialize(repodata['packages'][artifact_name]))


    # if os.path.exists(REG__REPODATA_SAMPLE_TEMP_FNAME):
    #     os.remove(REG__REPODATA_SAMPLE_TEMP_FNAME)






# -*- coding: utf-8 -*-

""" tests.test_common

(Mostly) unit tests for conda-content-trust/conda_content_trust/common.py.

Run the tests this way:
    pytest tests/test_common.py

"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import pytest

from conda_content_trust.common import *

# A 40-hex-character GPG public key fingerprint
SAMPLE_FINGERPRINT = 'f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589'
SAMPLE_UNKNOWN_FINGERPRINT = '0123456789abcdef0123456789abcdef01234567'

# The real key value of the public key (q, 32-byte ed25519 public key val),
# as a length-64 hex string.
SAMPLE_KEYVAL = 'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'

SAMPLE_GPG_KEY_OBJ = {
  'creation_time': 1571411344,
  'hashes': ['pgp+SHA2'],
  'keyid': SAMPLE_FINGERPRINT,
  'keyval': {
    'private': '',
    'public': {'q': SAMPLE_KEYVAL}
  },
  'method': 'pgp+eddsa-ed25519',
  'type': 'eddsa'
}

SAMPLE_ROOT_MD_CONTENT = {
  'delegations': {
    'key_mgr.json': {'pubkeys': [], 'threshold': 1},
    'root.json': {
      'pubkeys': ['bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07'],
      'threshold': 1}
  },
  'expiration': '2020-12-09T17:20:19Z',
  'metadata_spec_version': '0.1.0',  # TODO ✅⚠️❌💣: Update to 0.6.0 and remove the ".json" in the delegation names above, update the pubkey, and then re-sign this test metadata with the updated pubkey and adjust SAMPLE_GPG_SIG
  'type': 'root',
  'version': 1
}

SAMPLE_GPG_SIG = {
  'see_also': 'f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589',
  'other_headers': '04001608001d162104f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd58905025defd3d3',
  'signature': 'd6a3754dbd604a703434058c70db6a510b84a571236155df0b1f7f42605eb9e0faabca111d6ee808a7fcba663eafb5d66ecdfd33bd632df016fde3aed0f75201'
}

SAMPLE_SIGNED_ROOT_MD = {
  'signatures': {
    'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07': SAMPLE_GPG_SIG
  },
  'signed': SAMPLE_ROOT_MD_CONTENT
}

EXPECTED_SERIALIZED_SAMPLE_SIGNED_ROOT_MD = (b'{\n  '
    b'"signatures": {\n    '
        b'"bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07": {\n      '
            b'"other_headers": "04001608001d162104f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd58905025defd3d3",\n      '
            b'"see_also": "f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589",\n      '
            b'"signature": "d6a3754dbd604a703434058c70db6a510b84a571236155df0b1f7f42605eb9e0faabca111d6ee808a7fcba663eafb5d66ecdfd33bd632df016fde3aed0f75201"\n    }\n  },\n  '
    b'"signed": {\n    '
        b'"delegations": {\n      '
            b'"key_mgr.json": {\n        '
                b'"pubkeys": [],\n        '
                b'"threshold": 1\n      },\n      '
            b'"root.json": {\n        '
                b'"pubkeys": [\n          "bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07"\n        ],\n        '
                b'"threshold": 1\n      }\n    },\n    '
                b'"expiration": "2020-12-09T17:20:19Z",\n    '
                b'"metadata_spec_version": "0.1.0",\n    '  # TODO ✅⚠️❌💣: Update to 0.6.0 and remove the ".json" in the delegation names above, update the pubkey, and then re-sign this test metadata with the updated pubkey and adjust SAMPLE_GPG_SIG
                b'"type": "root",\n    '
                b'"version": 1\n  }\n}')

# # Some REGRESSION test data.
# REG__KEYPAIR_NAME = 'keytest_old'
REG__PRIVATE_BYTES = b'\xc9\xc2\x06\r~\r\x93al&T\x84\x0bI\x83\xd0\x02!\xd8\xb6\xb6\x9c\x85\x01\x07\xdat\xb4!h\xf97'
REG__PUBLIC_BYTES = b"\x01=\xddqIb\x86m\x12\xba[\xae'?\x14\xd4\x8c\x89\xcf\x07s\xde\xe2\xdb\xf6\xd4V\x1eR\x1c\x83\xf7"
REG__PUBLIC_HEX = '013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7'
REG__PRIVATE_HEX = 'c9c2060d7e0d93616c2654840b4983d00221d8b6b69c850107da74b42168f937'
# REG__MESSAGE_THAT_WAS_SIGNED = b'123456\x067890'
# # Signature is over REG__MESSAGE_THAT_WAS_SIGNED using key REG__PRIVATE_BYTES.
# REG__SIGNATURE = b'\xb6\xda\x14\xa1\xedU\x9e\xbf\x01\xb3\xa9\x18\xc9\xb8\xbd\xccFM@\x87\x99\xe8\x98\x84C\xe4}9;\xa4\xe5\xfd\xcf\xdaau\x04\xf5\xcc\xc0\xe7O\x0f\xf0F\x91\xd3\xb8"\x7fD\x1dO)*\x1f?\xd7&\xd6\xd3\x1f\r\x0e'
REG__HASHED_VAL = b'string to hash\n'
REG__HASH_HEX = '73aec9a93f4beb41a9bad14b9d1398f60e78ccefd97e4eb7d3cf26ba71dbe0ce'
# #REG__HASH_BYTES = b's\xae\xc9\xa9?K\xebA\xa9\xba\xd1K\x9d\x13\x98\xf6\x0ex\xcc\xef\xd9~N\xb7\xd3\xcf&\xbaq\xdb\xe0\xce'



# def test_sha512256():
#     # Test the SHA-512-truncate-256 hashing function w/ an expected result.
#     assert sha512256(REG__HASHED_VAL) == REG__HASH_HEX

#     # TODO: Test more?  Unusual input



def test_canonserialize():

    # Simple primitives
    assert canonserialize('') == b'""'
    assert canonserialize('a') == b'"a"'
    assert canonserialize(12) == b'12'
    assert canonserialize(SAMPLE_KEYVAL) == (
            b'"bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07"')

    # Tuples
    assert canonserialize((1, 2, 3)) == b'[\n  1,\n  2,\n  3\n]'
    assert canonserialize(('ABC', 1, 16)) == b'[\n  "ABC",\n  1,\n  16\n]'

    # Dictionaries indexed by ints, strings, or both
    assert canonserialize({}) == b'{}'
    assert canonserialize({1: 'v1', 2: 'v2'}) == (
            b'{\n  "1": "v1",\n  "2": "v2"\n}')
    assert canonserialize({'a': 'v1', 'b': 'v2'}) == (
            b'{\n  "a": "v1",\n  "b": "v2"\n}')
    with pytest.raises(TypeError):
        # Currently, json.dumps(...sort_keys=True) raises a TypeError while
        # sorting if it has to sort string keys and integer keys together.
        canonserialize({5: 'value', 'key': 9, 'key2': 'value2'})

    assert canonserialize(SAMPLE_SIGNED_ROOT_MD) == (
            EXPECTED_SERIALIZED_SAMPLE_SIGNED_ROOT_MD)

    # TODO: Tricksy tests that mess with encoding.



def test_keyfile_operations():
    """
    Unit tests for functions:
        keyfiles_to_keys
        keyfiles_to_bytes
    """
    # Test keyfiles_to_keys and keyfiles_to_bytes
    # Regression: load old key pair, two ways.
    # First, dump them to temp files (to test keyfiles_to_keys).
    with open('keytest_old.pri', 'wb') as fobj:
        fobj.write(REG__PRIVATE_BYTES)
    with open('keytest_old.pub', 'wb') as fobj:
        fobj.write(REG__PUBLIC_BYTES)
    loaded_old_private_bytes, loaded_old_public_bytes = keyfiles_to_bytes(
            'keytest_old')
    loaded_old_private, loaded_old_public = keyfiles_to_keys('keytest_old')

    # Clean up a bit.
    for fname in ['keytest_old.pri', 'keytest_old.pub']:
        if os.path.exists(fname):
            os.remove(fname)

    # Check the keys we wrote and then loaded.
    assert loaded_old_private_bytes == REG__PRIVATE_BYTES
    assert loaded_old_public_bytes == REG__PUBLIC_BYTES



def test_key_functions():
    # """
    # Tests for functions:
    #     from_bytes
    #     from_hex
    #     to_bytes
    #     to_hex
    #     is_equivalent_to
    # """
    # First key, generated in two ways:
    private_1_byt = PrivateKey.from_bytes(b'1'*32)
    private_1_hex = PrivateKey.from_hex('31'*32)  # hex representation of b'1'

    # Second key, generated in two ways:
    private_2_byt = PrivateKey.from_bytes(b'10' + b'1'*30)
    private_2_hex = PrivateKey.from_hex('3130' + '31'*30)

    # Regression key, generated in two ways:
    private_reg_byt = PrivateKey.from_bytes(REG__PRIVATE_BYTES)
    private_reg_hex = PrivateKey.from_hex(REG__PRIVATE_HEX)


    # Check these against each other and also against the expected output:
    #   - to_bytes
    #   - to_hex
    #   - is_equivalent_to
    #   - from_bytes
    #   - from_hex

    # key 1 from bytes vs key 1 from hex, also vs raw key 1 value
    assert private_1_byt.is_equivalent_to(private_1_hex)
    assert private_1_hex.is_equivalent_to(private_1_byt)
    assert b'1'*32 == private_1_byt.to_bytes() == private_1_hex.to_bytes()
    assert '31'*32 == private_1_byt.to_hex() == private_1_hex.to_hex()

    # key 1 vs key 2 vs regression key
    assert not private_1_byt.is_equivalent_to(private_2_byt)
    assert not private_2_byt.is_equivalent_to(private_1_byt)
    assert not private_1_byt.is_equivalent_to(private_2_hex)
    assert not private_2_hex.is_equivalent_to(private_1_byt)
    assert not private_reg_byt.is_equivalent_to(private_1_byt)
    assert not private_1_byt.is_equivalent_to(private_reg_byt)

    # key 2 from bytes vs key 2 from hex, also vs raw key 2 value
    assert private_2_byt.is_equivalent_to(private_2_hex)
    assert private_2_hex.is_equivalent_to(private_2_byt)
    assert b'10' + b'1'*30 == private_2_byt.to_bytes() == private_2_hex.to_bytes()
    assert '3130' + '31'*30 == private_2_byt.to_hex() == private_2_hex.to_hex()

    # regression key from bytes vs from hex, and vs raw key value
    assert private_reg_byt.is_equivalent_to(private_reg_hex)
    assert private_reg_hex.is_equivalent_to(private_reg_byt)
    assert REG__PRIVATE_BYTES == private_reg_byt.to_bytes()
    assert REG__PRIVATE_BYTES == private_reg_hex.to_bytes()
    assert REG__PRIVATE_HEX == private_reg_byt.to_hex()
    assert REG__PRIVATE_HEX == private_reg_hex.to_hex()


    # Test the behavior when is_equivalent_to is provided a bad argument.
    for bad_argument in ['1', 1, '1'*32, REG__PRIVATE_BYTES, b'1'*31, b'1'*33]:
        with pytest.raises(TypeError):
            private_reg_byt.is_equivalent_to(bad_argument)



# This is the version of the tests before PrivateKey and PublicKey classes were
# created to cut down on the utility function noise and make things easier to
# work with.
# def test_key_functions():
#     """
#     Unit tests for functions:
#         keyfiles_to_keys
#         keyfiles_to_bytes
#         key_to_bytes
#         public_key_from_bytes
#         private_key_from_bytes
#         keys_are_equivalent
#     """
#
#     # Test keyfiles_to_keys and keyfiles_to_bytes
#     # Regression: load old key pair, two ways.
#     # First, dump them to temp files (to test keyfiles_to_keys).
#     with open('keytest_old.pri', 'wb') as fobj:
#         fobj.write(REG__PRIVATE_BYTES)
#     with open('keytest_old.pub', 'wb') as fobj:
#         fobj.write(REG__PUBLIC_BYTES)
#     loaded_old_private_bytes, loaded_old_public_bytes = keyfiles_to_bytes(
#             'keytest_old')
#     loaded_old_private, loaded_old_public = keyfiles_to_keys('keytest_old')
#
#     # Clean up a bit.
#     for fname in ['keytest_old.pri', 'keytest_old.pub']:
#         if os.path.exists(fname):
#             os.remove(fname)
#
#     # Check the keys we wrote and then loaded.
#     assert loaded_old_private_bytes == REG__PRIVATE_BYTES
#     assert loaded_old_public_bytes == REG__PUBLIC_BYTES
#
#     # Test key object construction (could also call it "key loading")
#     other_private = private_key_from_bytes(b'1'*32)
#     other_private_dupe = private_key_from_bytes(b'1'*32)
#     other_public = public_key_from_bytes(b'2'*32)
#     other_public_dupe = public_key_from_bytes(b'2'*32)
#     for bad_argument in ['1', 1, '1'*32, loaded_old_private, b'1'*31, b'1'*33]:
#         with pytest.raises((TypeError, ValueError)):
#             public_key_from_bytes(bad_argument)
#         with pytest.raises((TypeError, ValueError)):
#             private_key_from_bytes(bad_argument)
#
#     # Test key equivalence checker.
#     assert keys_are_equivalent(other_private, other_private_dupe)
#     assert keys_are_equivalent(other_public, other_public_dupe)
#     assert not keys_are_equivalent(other_private, other_public)
#     assert not keys_are_equivalent(loaded_old_private, loaded_old_public)
#     assert not keys_are_equivalent(loaded_old_private, other_private)
#
#     for bad_argument in ['1', 1, '1'*32, REG__PRIVATE_BYTES, b'1'*31, b'1'*33]:
#         with pytest.raises(TypeError):
#             keys_are_equivalent(bad_argument, loaded_old_private)
#         with pytest.raises(TypeError):
#             keys_are_equivalent(loaded_old_private, bad_argument)
#         with pytest.raises(TypeError):
#             keys_are_equivalent(bad_argument, bad_argument)
#
#
#     # Test key_to_bytes
#     assert REG__PUBLIC_BYTES == key_to_bytes(loaded_old_public)
#     assert REG__PRIVATE_BYTES == key_to_bytes(loaded_old_private)
#     for bad_argument in ['1', 1, '1'*32, REG__PRIVATE_BYTES, b'1'*32]:
#         with pytest.raises(TypeError):
#             key_to_bytes(bad_argument)
#
#
#     # # Make a new keypair.  Returns keys and writes keys to disk.
#     # # Then load it from disk and compare that to the return value.  Exercise
#     # # some of the functions redundantly.
#     # assert keys_are_equivalent(generated_public, loaded_new_public)
#     # assert keys_are_equivalent(
#     #         loaded_new_private,
#     #         private_key_from_bytes(loaded_new_private_bytes))
#     # assert keys_are_equivalent(
#     #         loaded_new_public, public_key_from_bytes(loaded_new_public_bytes))



# Pull these from the integration tests in test_authentication.py


def test_is_gpg_signature():
    """
    Tests:
        is_gpg_signature
        checkformat_gpg_signature
        is_a_signature (only for cases relevant to gpg signatures)
        checkformat_signature (only for cases relevant to gpg signatures)
    """

    def expect_success(sig):
        checkformat_gpg_signature(sig)
        checkformat_signature(sig)
        assert is_gpg_signature(sig)
        assert is_a_signature(sig)

    def expect_failure(sig, exception_class):
        with pytest.raises(exception_class):
            checkformat_gpg_signature(sig)
        with pytest.raises(exception_class):
            checkformat_signature(sig)
        assert not is_gpg_signature(sig)
        assert not is_a_signature(sig)

    gpg_sig = {
            'other_headers': '04001608001d162104f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd58905025defd3d3',
            'signature': 'd6a3754dbd604a703434058c70db6a510b84a571236155df0b1f7f42605eb9e0faabca111d6ee808a7fcba663eafb5d66ecdfd33bd632df016fde3aed0f75201'
    }

    expect_success(gpg_sig)

    # Add optional fingerprint entry.
    gpg_sig['see_also'] = 'f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd589'
    expect_success(gpg_sig)

    # Too short
    gpg_sig['see_also'] = gpg_sig['see_also'][:-1]
    expect_failure(gpg_sig, ValueError)

    # Nonsense
    expect_failure(42, TypeError)

    del gpg_sig['see_also']

    # Also too short
    gpg_sig['signature'] = gpg_sig['signature'][:-1]
    expect_failure(gpg_sig, ValueError)








# def test_wrap_as_signable():
#     raise(NotImplementedError())

# def test_is_a_signable():
#     raise(NotImplementedError())

# def test_is_hex_signature():
#     raise(NotImplementedError())

def test_is_hex_key():
    assert is_hex_key('00' * 32)
    assert is_hex_key(SAMPLE_KEYVAL)
    assert not is_hex_key('00' * 31)
    assert not is_hex_key('00' * 33)
    assert not is_hex_key('00' * 64)
    assert not is_hex_key('1g' * 32)
    assert not is_hex_key(b'1g' * 32)

    pubkey_bytes = binascii.unhexlify(SAMPLE_KEYVAL)
    assert not is_hex_key(pubkey_bytes)

    public = PublicKey.from_bytes(pubkey_bytes)
    assert not is_hex_key(public)
    assert is_hex_key(public.to_hex())

def test_checkformat_hex_string():
    # TODO ✅: Add other tests.
    with pytest.raises(ValueError):
        checkformat_hex_string('A') # single case is important
    checkformat_hex_string('a')
    checkformat_hex_string(SAMPLE_KEYVAL)

# def test_checkformat_hex_key():
#     raise NotImplementedError()

# def test_checkformat_list_of_hex_keys():
#     raise NotImplementedError()

# def test_checkformat_byteslike():
#     raise NotImplementedError()

# def test_checkformat_natural_int():
#     raise NotImplementedError()

# def test_checkformat_expiration_distance():
#     raise NotImplementedError()

# def test_checkformat_utc_isoformat():
#     raise NotImplementedError()

# def test_checkformat_gpg_fingerprint():
#     raise NotImplementedError()

# def test_checkformat_gpg_signature():
#     raise NotImplementedError()


def test_checkformat_delegation():
    # TODO ✅: Add other tests.
    with pytest.raises(TypeError):
        checkformat_delegation(1)
    with pytest.raises(ValueError):
        checkformat_delegation({})
    with pytest.raises(ValueError):
        checkformat_delegation({
            'threshold': 0, 'pubkeys': ['01'*32]})
    with pytest.raises(ValueError):
        checkformat_delegation({
            'threshold': 1.5, 'pubkeys': ['01'*32]})
    checkformat_delegation({
        'threshold': 1, 'pubkeys': ['01'*32]})

    with pytest.raises(ValueError):
        checkformat_delegation({
            'threshold': 1, 'pubkeys': ['01'*31]})

    with pytest.raises(ValueError):
        checkformat_delegation({
            'threshold': 1, 'pubkeys': ['01'*31]})



def test_checkformat_delegating_metadata():

    checkformat_delegating_metadata(SAMPLE_SIGNED_ROOT_MD)
    # TODO ✅: Add a few other kinds of valid metadata to this test:
    #           - key_mgr metadata:
    #               - one signed using raw ed25519, one signed using OpenPGP
    #               - one with and one without version provided
    #           - root metadata:
    #               - one signed using raw ed25519 instead of OpenPGP


    for badval in [
            SAMPLE_ROOT_MD_CONTENT,
            # TODO ✅: Add more bad values (bad sig formats, etc.)
            ]:
        with pytest.raises( (TypeError, ValueError) ):
            checkformat_delegating_metadata(badval)




# def test_iso8601_time_plus_delta():
#     raise NotImplementedError()

# def test_is_hex_string():
#     raise(NotImplementedError())

# def test_set_expiry():
#     raise NotImplementedError()

